import os
from pathlib import Path


class WorkingDirectory(object):
    def __init__(self, cfg):
        self._cfg = cfg

        self._cwd = Path(os.getcwd())

        self._gitignore = self._cwd / ".gitignore"
        self._tf_config = self._cwd / "_eco_override.tf"

        self._terraform_folder = self._cwd / ".terraform"
        self._tf_environment = self._terraform_folder / "environment"

    def has_git_ignore(self):
        return self._gitignore.exists() and \
               self._gitignore.is_file()

    def create_git_ignore(self):
        with self._gitignore.open('w') as fp:
            fp.write("""_*tf
_*tfvars
""")

    def has_tf_files(self):
        return self._tf_config.exists() and \
               self._tf_config.is_file()

    def create_tf_files(self, **cli_args):

        composer = self._cfg.composer(**cli_args)

        if self._tf_config.exists():
            self._tf_config.chmod(0o666)

        with self._tf_config.open('w') as fp:
            composer.compose(fp)
        self._tf_config.chmod(0o444)

        if not self._terraform_folder.exists():
            self._terraform_folder.mkdir(0o777)

        with self._tf_environment.open('w') as fp:
            composer.compose_env(fp)

        if self._tf_environment.exists():
            self._tf_environment.chmod(0o666)



import click

from tfeco.configuration import ConfigurationFile
from tfeco.workingdirectory import WorkingDirectory

cfg = ConfigurationFile().load().save()


@click.group()
# @click.option('--debug/--no-debug', default=False)
def cli():
    """tf-eco"""
    pass


@cli.command('install-provider')
def install_provider_cmd():
    click.echo('install-provider')


@cli.command('init-stack')
def init_stack_cmd(*args, **kwargs):
    cwd = WorkingDirectory(cfg)
    if not cwd.has_git_ignore():
        cwd.create_git_ignore()

    cwd.create_tf_files(**kwargs)


# Add options from config files
for state, is_optional in cfg.facets():
    if is_optional:
        opt_func = click.option('--' + state, default="")
    else:
        opt_func = click.option('--' + state, required=True)
    init_stack_cmd = opt_func(init_stack_cmd)

if __name__ == '__main__':
    cli(obj={})



from functools import reduce

from tfeco.configuration_schema import ConfigSchema


class Composer(object):
    TAB_SPACES = 4

    def __init__(self, data, **facets):
        """
        :type facets: dict
        :type data: dict
        """
        self._data = data
        self._facets = facets

        ConfigSchema().load(self._data)

    def compose(self, fp):
        self._compose_backends(fp)
        self._compose_mappings(fp)
        self._compose_facets(fp)
        self._compose_providers(fp)

    def compose_env(self, fp):
        if self._data['facets']:
            if self._data['facets']['environment']:
                fp.write(
                    self._facets[
                        self._data['facets']['environment']
                    ]
                )

    def _compose_backends(self, fp):
        if 'backend' not in self._data:
            return
        backend_type = list(self._data['backend'].keys())[0]
        backend = self._data['backend'][backend_type]

        if 'key' in backend:
            backend['key'] = self._compose_backends_key()
            if backend['key'] is None:
                del backend['key']
            else:
                backend['key'] += '/state.tfstate'

        fp.write(
            """terraform {{\n    backend "{btype}" {items}}}""".format(
                btype=backend_type,
                items=self._build_block(backend, 1)
            ))
        fp.write("\n\n")

    def _compose_mappings(self, fp):
        mappings = self._data['mappings'] if 'mappings' in self._data else {}
        self._compose_locals(mappings, fp)

    def _compose_locals(self, locals, fp):
        keys = list(locals.keys())
        if len(keys):
            fp.write("locals {\n")
            mappings = []
            keys.sort()
            for key in keys:
                mappings.append("""    {name} = {keys}""".format(
                    name=key,
                    keys=self._build_block(locals[key], 1)
                ))
            fp.write("\n".join(mappings))
            fp.write("}\n\n")

    def _compose_facets(self, fp):
        facets = self._data['facets'] if 'facets' in self._data else {
            'state': [],
            'optional': []
        }
        state_facets = facets['state'] if 'state' in facets else []

        if 'composite' in facets:
            for composite_key, composite_parts in facets['composite'].items():
                # state_facets.remove(composite_key)
                for part in composite_parts:
                    state_facets.append(part)
        state_facets = list(set(state_facets))

        state_facets.sort()
        optional_facets = facets['optional'] if 'optional' in facets else []
        optional_facets.sort()

        for facet in state_facets:
            fp.write("variable \"{name}\" {{".format(name=facet))
            if facet in self._facets:
                fp.write("\n    default = \"" + self._facets[facet] + "\"\n")
            elif facet in optional_facets:
                fp.write("\n    default = \"\"\n")
            fp.write("}\n\n")

    def _build_config_line(self, kv, indent=1, padding_width=1):
        key_padding = '{:' + str(padding_width) + '}'

        if type(kv[1]) == bool:
            key = "true" if kv[1] else "false"
        else:
            key = "\"" + str(kv[1]) + "\""

        return "{indent}{key} = {value}".format(
            indent=" " * indent * self.TAB_SPACES,
            key=key_padding.format(kv[0]),
            value=key
        )

    def _build_block(self, configuration, indent=0):

        max_key_length = reduce(
            lambda key_length, key: max(key_length, len(key)),
            configuration.keys(),
            0
        )

        space_indent = " " * indent * self.TAB_SPACES
        block = "{\n"
        config_keys = list(configuration.keys())
        config_keys.sort()

        for key in config_keys:
            value = configuration[key]
            if type(value) == dict:
                key_padding = '{:' + str(max_key_length) + '}'
                block += " " * (indent + 1) * self.TAB_SPACES
                block += key_padding.format(key) + ' = '
                block += self._build_block(value, indent + 1)
            else:
                block += self._build_config_line(
                    [key, value],
                    indent + 1, max_key_length
                ) + "\n"

        block += space_indent + "}\n"

        return block

    def _compose_backends_key(self):

        facets = self._data['facets'] if 'facets' in self._data else {
            'state': []
        }

        state_facets = facets['state']

        ignore_facets = []

        composite_state_facets = []
        if 'composite' in facets:
            for composite_key, composite_parts in facets['composite'].items():
                if composite_key in state_facets:
                    state_facets.remove(composite_key)
                component_values = []
                for part in composite_parts:
                    ignore_facets.append(part)
                    state_facets.append(part)
                    component_values.append(self._facets[part])
                composite_state_facets.append(
                    composite_key + '=' + '/'.join(component_values)
                )

        filtered_state_facets = []
        for facet in state_facets:
            if facet in self._facets:
                if facet not in ignore_facets:
                    value = self._facets[facet]
                    if value != "":
                        filtered_state_facets.append(facet + '=' + value)

        if len(filtered_state_facets) == 0:
            return None

        filtered_state_facets.sort()

        return '/'.join(filtered_state_facets + composite_state_facets)

    def _compose_providers(self, fp):
        providers = self._data['providers'] \
            if 'providers' in self._data \
            else {}
        provider_names = list(providers.keys())
        provider_names.sort()

        for name in provider_names:
            configurations = providers[name]
            for provider_instance in configurations:
                self._compose_provider(name, provider_instance, fp)
                # fp.write("\n\n")

    def _compose_provider(self, name, provider_instance, fp):
        fp.write("provider \"" + name + "\" ")

        config_keys = list(provider_instance.keys())
        config_keys.sort()

        fp.write(self._build_block(provider_instance, 0) + "\n")



from pathlib import Path

import pyaml
import yaml

from tfeco.composer import Composer
from tfeco.configuration_schema import ConfigSchema


class ConfigurationFile(object):
    """
    Describes a file path to the configuration for the utility
    """

    def __init__(self):
        self._config_folder = Path(Path.home(), '.config', 'tf-ecosystem')
        self._config_file = Path(self._config_folder, 'config.ini')
        self._config = {}

    def load(self) -> 'ConfigurationFile':
        if not self._config_folder.exists():
            self._config_folder.mkdir()

        self._init_defaults()

        if self._config_file.exists():
            with self._config_file.open('r') as fp:
                self._config = yaml.load(fp, Loader=yaml.CLoader)
                ConfigSchema().load(self._config)

        return self

    def save(self) -> 'ConfigurationFile':
        with self._config_file.open('w') as fp:
            pyaml.dump(self._config, fp)

        return self

    def facets(self):
        states = self._config['facets']['state']
        optional = self._config['facets']['optional']

        for state in states:
            yield state, state in optional

    def composer(self, **facets) -> Composer:
        return Composer(self._config, **facets)

    def _init_defaults(self):
        if self._config is None or len(self._config) == 0:
            self._config = {
                'facets': {
                    'state': [
                        'account',
                        'region',
                        'environment',
                        'stack'
                    ],
                    'optional': ['account']
                },
                'backend': {
                    's3': {
                        'bucket': 'my-bucket',
                        'region': 'us-east-1',
                        'acl': 'bucket_acl',
                        'dynamodb_table': 'dynamodb_table',
                        'role_arn': 'role_arn',
                        'key': 'bucket_key'
                    }
                },
                'mappings': {
                    'account-names': {
                        'test': '012345678901',
                        'stage': '234567890123',
                        'live': '456789012345'
                    }
                },
                'providers': {
                    'aws': []
                }
            }



from unittest import TestCase

from io import StringIO

from tfeco.composer import Composer
from tfeco.configuration import ConfigurationFile


class TestComposer(TestCase):

    def test__compose_defaults(self):
        cfg = ConfigurationFile()
        cfg._init_defaults()
        c = Composer(cfg._config)

        mockIO = StringIO()

        c.compose(mockIO)

        mockIO.seek(0)

        self.assertEqual("""terraform {
    backend "s3" {
        acl            = "bucket_acl"
        bucket         = "my-bucket"
        dynamodb_table = "dynamodb_table"
        region         = "us-east-1"
        role_arn       = "role_arn"
    }
}

locals {
    account-names = {
        live  = "456789012345"
        stage = "234567890123"
        test  = "012345678901"
    }
}

variable "account" {
    default = ""
}

variable "environment" {}

variable "region" {}

variable "stack" {}

""", mockIO.read())

    def test__compose_backends(self):
        c = Composer({
            'backend': {
                's3': {
                    'bucket': 'my-bucket',
                    'region': 'bucket_region',
                    'acl': 'bucket_acl',
                    'dynamodb_table': 'dynamodb_table',
                    'role_arn': 'role_arn',
                    'key': 'bucket_key'
                }
            },
            'providers': {}
        })

        mockIO = StringIO()

        c.compose(mockIO)

        mockIO.seek(0)

        self.assertEqual("""terraform {
    backend "s3" {
        acl            = "bucket_acl"
        bucket         = "my-bucket"
        dynamodb_table = "dynamodb_table"
        region         = "bucket_region"
        role_arn       = "role_arn"
    }
}

""", mockIO.read())

    def test__compose_mappings(self):
        c = Composer({
            'mappings': {
                'account-names': {
                    'test': '012345678901',
                    'stage': '234567890123',
                    'live': '456789012345'
                },
                'randoms': {
                    'foo': 'bar',
                    'hello': 'world'
                }
            },
            'providers': {}
        })

        mockIO = StringIO()

        c.compose(mockIO)

        mockIO.seek(0)

        self.assertEqual("""locals {
    account-names = {
        live  = "456789012345"
        stage = "234567890123"
        test  = "012345678901"
    }

    randoms = {
        foo   = "bar"
        hello = "world"
    }
}

""", mockIO.read())

    def test__compose_facets(self):
        c = Composer({
            'facets': {
                'state': ['hello', 'world', 'foo', 'bar'],
                'optional': ['foo', 'bar']
            },
            'providers': {}
        })

        mockIO = StringIO()

        c.compose(mockIO)

        mockIO.seek(0)

        self.assertEqual("""variable "bar" {
    default = ""
}

variable "foo" {
    default = ""
}

variable "hello" {}

variable "world" {}

""", mockIO.read())

    def test__compose_backends_key(self):
        c = Composer({
            'facets': {
                'state': ['test_key', 'missing_key']
            },
        }, test_key='mock')

        key_path = c._compose_backends_key()

        self.assertEqual(key_path, 'test_key=mock')

    def test__compose_providers(self):
        c = Composer({
            'providers': {
                'aws': [{
                    "region": "${var.region}",
                    "skip_get_ec2_platforms": True,
                    "assume_role": {
                        "role_arn": "arn:aws:iam::${lookup("
                                    "var.account-names, "
                                    "var.account"
                                    ")}:role/${var.environment}"
                                    "-terraform-provisioner",
                        "session_name": "${terraform.env}-${replace("
                                        "var.stack,"
                                        "\"/[^\\w\\d-]/\",\"\")}"
                                        "-terraform"
                    }
                }]
            }
        })

        mockIO = StringIO()

        c.compose(mockIO)

        mockIO.seek(0)

        self.assertEqual("""provider "aws" {
    assume_role            = {
        role_arn     = "arn:aws:iam::${lookup(var.account-names, var.account)}:role/${var.environment}-terraform-provisioner"
        session_name = "${terraform.env}-${replace(var.stack,"/[^\\w\\d-]/","")}-terraform"
    }
    region                 = "${var.region}"
    skip_get_ec2_platforms = true
}

""", mockIO.read())

    def test__compose_block(self):
        c = Composer({})
        block = c._build_block({
            'one': 'foo',
            'two': 'bar',
            'three': 'hello',
            'four': 'world'
        })

        self.assertEqual("""{
    four  = "world"
    one   = "foo"
    three = "hello"
    two   = "bar"
}
""", block)

    def test__compose_block_dict(self):
        c = Composer({})
        block = c._build_block({
            'one': 'foo',
            'two': 'bar',
            'three': 'hello',
            'four': {
                'five': 'world',
                'six': 'the',
                'seven': 'world',
                'eight': 'is'
            }
        })

        self.assertEqual("""{
    four  = {
        eight = "is"
        five  = "world"
        seven = "world"
        six   = "the"
    }
    one   = "foo"
    three = "hello"
    two   = "bar"
}
""", block)

    def test__composite_facet(self):
        c = Composer({
            'backend': {
                's3': {
                    'bucket': 'my-bucket',
                    'region': 'us-east-1',
                    'acl': 'bucket_acl',
                    'dynamodb_table': 'dynamodb_table',
                    'role_arn': 'role_arn',
                    'key': 'bucket_key'
                }
            },
            'facets': {
                'state': ['env', 'comp_stack'],
                'composite': {
                    "comp_stack": ['stack', 'substack']
                }
            }
        }, env='live', stack='mock_stack', substack='mock_substack')

        mockIO = StringIO()

        c.compose(mockIO)

        mockIO.seek(0)

        self.assertEqual("""terraform {
    backend "s3" {
        acl            = "bucket_acl"
        bucket         = "my-bucket"
        dynamodb_table = "dynamodb_table"
        key            = "env=live/comp_stack=mock_stack/mock_substack/state.tfstate"
        region         = "us-east-1"
        role_arn       = "role_arn"
    }
}

variable "env" {
    default = "live"
}

variable "stack" {
    default = "mock_stack"
}

variable "substack" {
    default = "mock_substack"
}

""", mockIO.read())



from marshmallow import Schema, fields


class ConfigFacetSchema(Schema):
    state = fields.List(fields.String)
    optional = fields.List(fields.String)


class ConfigSchema(Schema):
    facets = fields.Nested(ConfigFacetSchema)
    backend = fields.Dict()
    mappings = fields.Dict()
    providers = fields.Dict()



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='tf-ecosystem',
    author='Drew J. Sonne',
    author_email='drew.sonne@gmail.com',
    description='A framework for initialising a terraform environment',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3',
    version='1.0.5',
    url='http://github.com/drewsonne/tf-ecosystem',
    license='LGPLv2',
    include_package_data=True,
    packages=find_packages(),
    keywords='terraform cli',
    install_requires=[
        'marshmallow',
        'click',
        'pyaml',
        'termcolor'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: '
        'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: System',
        'Topic :: System :: Systems Administration',
    ],
    setup_requires=['nose>=1.0'],
    entry_points={
        'console_scripts': [
            'tf-eco = tfeco.scripts:cli',
            'tf-ecosystem = tfeco.scripts:cli'
        ]
    },
    test_suite='nose.collector',
    tests_require=[
        'coverage',
        'nose',
        'nose-cover3'
    ],
    zip_safe=True,
)
